from .validator_testing_tools import ValidatorTestingTools
from tripoli import IIIFValidator
from tripoli.resource_validators.base_validator import BaseValidator
from tripoli.validator_logging import ValidatorLogError, ValidatorLogWarning, Path
from tripoli.exceptions import FailFastException


class TestBaseValidatorMixin(ValidatorTestingTools):

    def setUp(self):
        iv = IIIFValidator()
        iv.fail_fast = False
        iv.collect_errors = True
        iv.collect_warnings = True
        iv.debug = False
        self.test_subject = BaseValidator(iv)

    def test_compare_dicts(self):
        man = {"field_1": "value", "field_3": "value"}

        # Test that nothing is changed in valid case.
        schema = {"field_1": self.fake_valid, "field_2": self.fake_valid}
        val = self.test_subject._compare_dicts(schema, man)
        self.assertFalse(self.has_errors())
        self.assertEqual(man, val)
        self.clear_errors_and_warnings()

        # Test that warning is logged in invalid case.
        schema = {"field_1": self.fake_invalid, "field_2": self.fake_valid}
        val = self.test_subject._compare_dicts(schema, man)
        self.assertEqual(man, val)
        self.assertTrue(self.has_errors())
        self.clear_errors_and_warnings()

        # Test that making a change does not affect original argument.
        schema = {"field_1": self.fake_change}
        old_man = man.copy()
        val = self.test_subject._compare_dicts(schema, man)
        self.assertEqual(old_man, man)
        self.assertNotEqual(val, man)

    def check_recommended_fields(self):
        # Test a warning comes through when missing a recommended field.
        self.test_subject._check_recommended_fields("test", {}, ["test_field"])
        self.assertTrue(self.has_warnings())
        self.clear_errors_and_warnings()

        # Test no warning when all recommended fields are there.
        self.test_subject._check_recommended_fields("test", {"test_field": "value"}, ["test_field"])
        self.assertFalse(self.has_warnings())

    def check_unknown_fields(self):
        # Test a warning comes through when an unknown field is included.
        self.test_subject._check_unknown_fields("test", {"other_field": "value"}, ["test_field"])
        self.assertTrue(self.has_warnings())
        self.clear_errors_and_warnings()

        # Test no warning when all fields are known.
        self.test_subject._check_unknown_fields("test", {"test_field": "value"}, ["test_field"])
        self.assertFalse(self.has_warnings())

    def check_forbidden_fields(self):
        # Test an error comes through when forbidden field is included.
        self.test_subject._check_forbidden_fields("test", {"test_field": "value"}, ["test_field"])
        self.assertTrue(self.has_errors())
        self.clear_errors_and_warnings()

        # Test no error when no fields are forbidden.
        self.test_subject._check_forbidden_fields("test", {"other_field": "value"}, ["test_field"])
        self.assertFalse(self.has_errors())

    def check_required_fields(self):
        # Test an error comes through when a required field is missing.
        self.test_subject._check_required_fields("test", {"other_field": "value"}, ["test_field"])
        self.assertTrue(self.has_errors())
        self.clear_errors_and_warnings()

        # Test no error when all required fields are included.
        self.test_subject._check_required_fields("test", {"test_field": "value"}, ["test_field"])
        self.assertFalse(self.has_errors())

    def test_str_or_val_lang_type(self):
        """Allows stings, lists of strings, val/lang dicts, and lists of val/lang dicts"""
        valid_inputs = ['hello', ['hello', 'there'], {'@language': 'en', '@value': 'hello'},
                        [{'@language': 'en', '@value': 'hello'}, {'@language': 'fr', '@value': 'bonjour'}],
                        {'@value': 'Hello'}]
        self.assert_no_errors_with_inputs(self.test_subject._str_or_val_lang_type, valid_inputs)

        invalid_inputs = [0, {'@value': 14, '@language': 'en'}, ]
        self.assert_errors_with_inputs(self.test_subject._str_or_val_lang_type, invalid_inputs)

    def test_repeatable_str_type(self):
        """Allows strings or lists of strings."""
        valid_inputs = ['hello', ['hello', 'there']]
        self.assert_no_errors_with_inputs(self.test_subject._repeatable_string_type, valid_inputs)

    def test_repeatable_uri_type(self):
        """Allow a single uri or list of uris (in both uri formats)"""
        valid_inputs = ['http://google.ca', ['http://google.com', {'@id': 'http://bing.com'}]]
        self.assert_no_errors_with_inputs(self.test_subject._repeatable_uri_type, valid_inputs)

        invalid_inputs = ['hello', ['hello', 'everyone']]
        self.assert_errors_with_inputs(self.test_subject._repeatable_uri_type, invalid_inputs)

    def test_http_uri_type(self):
        """Allow a single uri in either format that must be http/s"""
        valid_inputs = ['http://google.ca', {'@id': 'https://bing.com'}]
        self.assert_no_errors_with_inputs(self.test_subject._http_uri_type, valid_inputs)

        invalid_inputs = ['ftp://google.ca', ['http://google.ca', 'http://bing.ca']]
        self.assert_errors_with_inputs(self.test_subject._http_uri_type, invalid_inputs)

    def test_uri_type(self):
        """Allow a single uri in either format."""
        valid_inputs = ['http://google.ca', 'ftp://google.ca', {'@id': 'http://google.ca'}]
        self.assert_no_errors_with_inputs(self.test_subject._uri_type, valid_inputs)

        invalid_inputs = [{'key': 'http:google.ca'}, 'hello', ['http://google.ca']]
        self.assert_errors_with_inputs(self.test_subject._uri_type, invalid_inputs)

    def test_metadata_field(self):
        """Allow properly formatted metadata as specified by the presentation api."""
        valid_inputs = [
            [{'value': 'hello', 'label': 'greeting'}],
            [{'value': [{'@language': 'en', '@value': 'hello'}], 'label': 'greeting'}],
            [{'value': 'hello', 'label': {'@language': 'en', '@value': 'greeting'}}],
        ]
        self.assert_no_errors_with_inputs(self.test_subject.metadata_field, valid_inputs)

        invalid_inputs = [
            [{'bad_key': 'wow', 'this_sucks': 'yup'}],
            ['a string'],
            "a stringier string",
            [{"value": "missing label"}]
        ]
        self.assert_errors_with_inputs(self.test_subject.metadata_field, invalid_inputs)

    def test_general_image_resource(self):
        """Test correct behaviour of general image resource validation."""
        val = "http://google.ca"
        self.test_subject._general_image_resource("unknown_field", val)
        self.assertTrue(self.has_warnings())
        self.assertFalse(self.has_errors())
        self.clear_errors_and_warnings()

        val = {"@id": "http://google.ca"}
        self.test_subject._general_image_resource("unknown_field", val)
        self.assertTrue(self.has_warnings())
        self.assertFalse(self.has_errors())
        self.clear_errors_and_warnings()

        val = {"@id": "http://google.ca",
               "@type": "dctypes:Image",
               "service": {"@context": "http://iiif.io/api/image/2/context.json",
                           "@id": "http://google.ca",
                           "profile": "http://google.ca"}}
        self.test_subject._general_image_resource("unknown_field", val)
        self.assertFalse(self.has_warnings())
        self.assertFalse(self.has_errors())

    def test_view_dir_hint_field(self):
        self.test_subject.VIEW_HINTS = {"paged", "non-paged"}
        self.test_subject.VIEW_DIRS = {"paged", "non-paged"}
        self.assert_no_errors_with_inputs(self.test_subject.viewing_hint_field, ["paged", "non-paged"])
        self.assert_errors_with_inputs(self.test_subject.viewing_hint_field, ["error"])
        self.assert_no_errors_with_inputs(self.test_subject.viewing_dir_field, ["paged", "non-paged"])
        self.assert_errors_with_inputs(self.test_subject.viewing_dir_field, ["error"])

    def test_html_validation(self):
        # Valid (path, field, html-value) tuples
        valid_inputs = {
            (Path(('metadata',)), 'label', "<a href='http://google.ca'>Google</a>"),
            (Path(('sequences', 0, 'metadata')), 'value', "<span><p>Test</p></span>"),
            (Path(tuple()), 'description', "<p><br/></p>"),
            (Path(tuple()), 'attribution', "<img src='fake source'/>")
        }

        for p, f, v in valid_inputs:
            self.test_subject._path = p
            self.test_subject._check_html(f, v)
            self.assertFalse(self.has_warnings_or_errors(), (p,v,f))

        # Invalid (path, field, html-value) tuples
        invalid_inputs = {
            (Path(tuple()), 'label', "<a href='http://google.ca'>Bad field</a>"),
            (Path(tuple()), 'description', "<script>Nasty scripting</script>"),
            (Path(tuple()), 'description', "<a target='something'>Bad attribute</a>"),
            (Path(tuple()), 'description', "this is <a>badly formatted.</a>"),
        }

        for p, f, v in invalid_inputs:
            self.test_subject._path = p
            self.test_subject._check_html(f, v)
            self.assertTrue(self.has_errors(), (p,f,v))
            self.clear_errors_and_warnings()

    def test_mute_error(self):
        # Test error is caught.
        val, err = self.test_subject.mute_errors(self.fake_invalid, "value")
        self.assertIn(ValidatorLogError('test error', ('fake field',)), err)
        self.assertFalse(self.has_errors())

        # Test fail fast is not triggered.
        self.test_subject._IIIFValidator.fail_fast = True
        try:
            val, err = self.test_subject.mute_errors(self.fake_invalid, "value")
        except FailFastException:
            self.fail("FailFastException was raised.")
        self.assertIn(ValidatorLogError('test error', ('fake field',)), err)
        self.assertFalse(self.has_errors())

    def test_error_to_warning(self):
        @self.test_subject.errors_to_warnings
        def log_error(self, value):
            self.log_error("fake field", "test error")

        # Test that error was converted to warning properly.
        log_error(self.test_subject, "value")
        self.assertFalse(self.has_errors())
        self.assertTrue(self.has_warnings())
        self.assertEqual(self.test_subject.warnings.pop(), ValidatorLogWarning('test error', ('fake field',)))

        # Test that fail fast is not triggered.
        self.test_subject._IIIFValidator.fail_fast = True
        try:
            log_error(self.test_subject, "value")
        except FailFastException:
            self.fail("FailFastException was raised.")
        self.assertTrue(self.has_warnings())
        self.assertFalse(self.has_errors())

    def test_warning_to_error(self):
        @self.test_subject.warnings_to_errors
        def log_warning(self, value):
            self.log_warning("fake field", "test error")

        # Test that warning is converted to error properly
        log_warning(self.test_subject, "value")
        self.assertTrue(self.has_errors())
        self.assertFalse(self.has_warnings())
        self.assertEqual(self.test_subject.errors.pop(), ValidatorLogError('test error', ('fake field',)))




class FailFastException(Exception):
    """Raised when validation should end because error is hit."""
    pass


class TypeParseException(Exception):
    """Raised when a `@type` can not be parsed from input data."""
    pass



import json
import os

from .validator_testing_tools import ValidatorTestingTools
from tripoli import IIIFValidator


class TestErrorCollection(ValidatorTestingTools):

    def setUp(self):
        self.test_subject = IIIFValidator()
        self.test_subject.logger.disabled = True
        self.base_dir = os.path.dirname(os.path.realpath(__file__))

    def test_error_collection(self):

        with open(os.path.join(self.base_dir, 'fixtures/error_collection.json')) as f:
            collection = json.load(f)

        for m in collection['manifests']:
            self.test_subject.validate(m['manifest'])
            self.assertTrue(self.has_errors(), "Did not catch error on '{}'".format(m['label']))
            self.assertEqual(str(self.test_subject.errors[0]), m['error'])
            self.clear_errors_and_warnings()



import unittest
import inspect

from tripoli.validator_logging import ValidatorLog


class ValidatorTestingTools(unittest.TestCase):

    def setUp(self):
        self.test_subject = None
        raise NotImplemented

    def has_error(self, error):
        """Check if a particular error was raised."""
        return error in self.test_subject._errors

    def has_errors(self):
        return bool(self.test_subject._errors)

    def has_warning(self, warning):
        return warning in self.test_subject._warnings

    def has_warnings(self):
        return bool(self.test_subject._warnings)

    def has_warnings_or_errors(self):
        return (bool(self.test_subject._warnings) or bool(self.test_subject._errors))

    def clear_errors(self):
        self.test_subject._errors = ValidatorLog(self.test_subject.unique_logging)

    def clear_warnings(self):
        self.test_subject._warnings = ValidatorLog(self.test_subject.unique_logging)

    def clear_errors_and_warnings(self):
        self.test_subject._errors = ValidatorLog(self.test_subject.unique_logging)
        self.test_subject._warnings = ValidatorLog(self.test_subject.unique_logging)

    def assert_no_errors_with_inputs(self, fn, inputs):
        """Run validation fn on each input, asserting no error is logged."""
        args = inspect.signature(fn).parameters
        for arg in inputs:
            val = fn('unknown_field', arg) if "field" in args else fn(arg)
            try:
                self.assertFalse(self.has_errors(), "{} is not valid input".format(arg))
                self.assertEqual(arg, val)
            finally:
                self.clear_errors()

    def assert_errors_with_inputs(self, fn, inputs):
        args = inspect.signature(fn).parameters
        for arg in inputs:
            val = fn('unknown_field', arg) if "field" in args else fn(arg)
            try:
                self.assertTrue(self.has_errors(), "{} is valid input".format(arg))
                self.assertEqual(arg, val)
            finally:
                self.clear_errors()

    def fake_valid(self, value):
        return value

    def fake_invalid(self, value):
        self.test_subject.log_error("fake field", "test error")
        return value

    def fake_change(self, value):
        return str(value) + "changed stuff"



from .canvas_validator import CanvasValidator
from .annotation_validator import AnnotationValidator
from .manifest_validator import ManifestValidator
from .sequence_validator import SequenceValidator
from .image_content_validator import ImageContentValidator
from .base_validator import BaseValidator



from .tripoli import IIIFValidator, AnnotationValidator, \
    ManifestValidator, CanvasValidator, SequenceValidator, ImageContentValidator,\
    __version__
from .resource_validators.base_validator import BaseValidator



# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import traceback
from itertools import zip_longest


class Path:
    """Class representing path within document."""
    def __init__(self, path=None):
        """ Create a Path.

        :param path: Tuple of strings and ints.
        """
        self._path = path if path is not None else tuple()
        self.__no_index_path = None

    def __eq__(self, other):
        return self.__no_index_path == other.__no_index_path

    def __len__(self):
        return len(self._path)

    def __str__(self):
        return ' @ data[%s]' % ']['.join(map(repr, self._path)) if self._path else ''

    def __repr__(self):
        return 'Path({})'.format(", ".join(repr(x) for x in self._path))

    def __iter__(self):
        return self._path.__iter__()

    def __add__(self, other):
        if isinstance(other, (str, int)):
            return Path(self._path + (other,))
        if isinstance(other, tuple):
            return Path(self._path + other)
        if isinstance(other, Path):
            return Path(self._path + other._path)
        raise NotImplemented

    def __hash__(self):
        return hash(self._path)

    @property
    def no_index_path(self):
        if self.__no_index_path is None:
            self.__no_index_path = tuple(filter(lambda x: isinstance(x, str), self._path))
        return self.__no_index_path

    @property
    def path(self):
        return self._path

    def no_index_eq(self, other):
        """Return true if paths are the same, ignoring indexes.

        Useful for hashing, as we typically only care about storing
        one instance of each error/warning per list.
        """
        return self.no_index_path == other.no_index_path

    def no_index_endswith(self, path):
        """Return true is self.no_index_path has ``path`` as a suffix.

        :param path: Either a str-tuple or a Path.
        """
        if isinstance(path, Path):
            path = path.no_index_path
        return path == self.no_index_path[-(len(path)):]


class ValidatorLog:
    """Log which provides unified interface for either set or list like behaviour."""
    def __init__(self, unique_logging=True):
        self.unique_logging = unique_logging
        self._entries = set() if unique_logging else []

    def add(self, log_entry):
        """Add an entry to the log.

        :param log_entry: A ValidatorLogEntry to add to self.
        """
        if isinstance(log_entry, ValidatorLogEntry):
            if self.unique_logging:
                self._entries.add(log_entry)
            else:
                self._entries.append(log_entry)
        else:
            raise TypeError('log_entry must be a ValidatorLogEntry')

    def update(self, log_entry):
        """Add all entries from log_entry to self.

        :param log_entry: A ValidatorLog to update from.
        """
        for entry in log_entry._entries:
            self.add(entry)

    def __iter__(self):
        return iter(self._entries)

    def __bool__(self):
        return bool(self._entries)


class ValidatorLogEntry:
    """Basic error logging class with comparison behavior for hashing."""

    def __init__(self, msg, path, tb=None):
        """

        :param msg: A message associated with the log entry.
        :param path: A tuple representing the path where entry was logged.
        :param tb: A traceback.extract_stack() list from the point entry was logged.
        """

        #: A message associated with the log entry.
        self.msg = msg

        #: A tuple representing the path where the entry was created.
        self.path = Path(path)

        self._tb = tb if tb else []

    def print_trace(self):
        """Print the stored traceback if it exists."""
        traceback.print_list(self._tb)

    def path_str(self):
        return str(self.path)

    def log_str(self):
        return self.msg + self.path_str()

    def __lt__(self, other):
        return len(self.path) < len(other.path)

    def __hash__(self):
        return hash(self.path.no_index_path) ^ hash(self.msg)

    def __eq__(self, other):
        return self.path.no_index_path == other.path.no_index_path\
                and self.msg == other.msg


class ValidatorLogWarning(ValidatorLogEntry):
    """Class to hold and present warnings."""

    def __str__(self):
        output = "Warning: {}".format(self.msg)
        return output + self.path_str()

    def __repr__(self):
        return "ValidatorLogWarning('{}', {})".format(self.msg, self.path)


class ValidatorLogError(ValidatorLogEntry):
    """Class to hold and present errors."""

    def __str__(self):
        output = "Error: {}".format(self.msg)
        return output + self.path_str()

    def __repr__(self):
        return "ValidatorLogError('{}', {})".format(self.msg, self.path)



# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from collections import OrderedDict

from .base_validator import BaseValidator


class CanvasValidator(BaseValidator):
    VIEW_HINTS = {'non-paged', 'facing-pages'}

    KNOWN_FIELDS = BaseValidator.COMMON_FIELDS | {"height", "width", "otherContent", "images"}
    FORBIDDEN_FIELDS = {"format", "viewingDirection", "navDate", "startCanvas", "first", "last", "total",
                        "next", "prev", "startIndex", "collections", "manifests", "members", "sequences",
                        "structures", "canvases", "resources", "ranges"}
    REQUIRED_FIELDS = {"label", "@id", "@type", "height", "width"}

    def __init__(self, iiif_validator):
        super().__init__(iiif_validator)
        self.CanvasSchema = OrderedDict((
            ('@id', self.id_field),
            ('@type', self.type_field),
            ('label', self.label_field),
            ('height', self.height_field),
            ('width', self.width_field),
            ('other_content', self.other_content_field),
            ('images', self.images_field)
        ))
        self.setup()

    def _run_validation(self, **kwargs):
        self._check_all_key_constraints("canvas", self._json)
        self.canvas_uri = self._json['@id']
        return self._compare_dicts(self.CanvasSchema, self._json)

    def _raise_additional_warnings(self, validation_results):
        # Canvas should have a thumbnail if it has multiple images.
        if len(validation_results.get('images', [])) > 1 and not validation_results.get("thumbnail"):
            self.log_warning("thumbnail", "Canvas SHOULD have a thumbnail when there is more than one image")

    def type_field(self, value):
        """Assert that ``@type == 'sc:Canvas``"""
        if value != "sc:Canvas":
            self.log_error("@type", "@type must be 'sc:Canvas'.")
        return value

    def images_field(self, value):
        """Validate ``images`` list.

        Calls a sub-validation procedure handled by the :class:`AnnotationValidator`.
        """
        if isinstance(value, list):
            path = self._path + ("images",)
            results = []
            for i, anno in enumerate(value):
                temp_path = path + i
                results.append(self._sub_validate(self.AnnotationValidator, anno, temp_path,
                               canvas_uri=self.canvas_uri))
            return results
        if not value:
            self.log_warning("images", "'images' SHOULD have values.")
            return value
        self.log_error("images", "'images' must be a list.")
        return value

    def other_content_field(self, value):
        """Validate ``otherContent`` field."""
        if not isinstance(value, list):
            self.log_error("otherContent", "otherContent must be a list.")
            return value
        return [self._uri_type("otherContent", item['@id']) for item in value]



# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from collections import OrderedDict

from .base_validator import BaseValidator


class SequenceValidator(BaseValidator):
    VIEW_DIRS = {'left-to-right', 'right-to-left',
                 'top-to-bottom', 'bottom-to-top'}
    VIEW_HINTS = {'individuals', 'paged', 'continuous'}

    KNOWN_FIELDS = BaseValidator.COMMON_FIELDS | {"viewingDirection", "startCanvas", "canvases"}
    FORBIDDEN_FIELDS = {"format", "height", "width", "navDate", "first", "last", "total", "next", "prev",
                        "startIndex", "collections", "manifests", "sequences", "structures", "resources",
                        "otherContent", "images", "ranges"}
    REQUIRED_FIELDS = {"@type", "canvases"}

    def __init__(self, iiif_validator):
        super().__init__(iiif_validator)
        self.emb = None
        self.EmbSequenceSchema = OrderedDict((
            ('@type', self.type_field),
            ('@context', self.context_field),
            ('@id', self.id_field),
            ('startCanvas', self.startCanvas_field),
            ('viewingDirection', self.viewing_dir_field),
            ('canvases', self.canvases_field),
        ))

        self.LinkedSequenceSchema = OrderedDict((
            ('@type', self.type_field),
            ('@id', self.id_field),
            ('canvases', self._not_allowed)
        ))
        self.setup()

    def _run_validation(self, **kwargs):
        self._check_all_key_constraints("sequence", self._json)
        return self._validate_sequence(**kwargs)

    def _validate_sequence(self, emb=True):
        self.emb = emb
        if self.emb:
            return self._compare_dicts(self.EmbSequenceSchema, self._json)
        else:
            return self._compare_dicts(self.LinkedSequenceSchema, self._json)

    def _raise_additional_warnings(self, validation_results):
        pass

    def type_field(self, value):
        """Assert that ``@type`` == ``sc:Sequence``"""
        if value != "sc:Sequence":
            self.log_error("@type", "@type must be 'sc:Sequence'")
        return value

    def context_field(self, value):
        """Assert that ``@context`` is the IIIF 2.0 presentation API if it is allowed."""
        if self.emb:
            self.log_error("@context", "@context field not allowed in embedded sequence.")
            return value
        if value != self.PRESENTATION_API_URI:
            self.log_error("@context", "unknown context.")
        return value

    def startCanvas_field(self, value):
        """Validate ``startCanvas`` field."""
        canvases = self._json.get('canvases', [])
        if any(True for can in canvases if can.get('@id') == value):
            pass
        else:
            self.log_error("startCanvas", "'startCanvas' must refer to the @id of some canvas in this sequence.")
        return value

    def canvases_field(self, value):
        """Validate ``canvases`` list for Sequence."""
        if not isinstance(value, list):
            self.log_error("canvases", "'canvases' MUST be a list.")
            return value
        if len(value) < 1:
            self.log_error("canvases", "'canvases' MUST have at least one entry")
            return value
        path = self._path + ("canvases",)
        results = []
        for i, canvas in enumerate(value):
            temp_path = path + i
            results.append(self._sub_validate(self.CanvasValidator, canvas, temp_path))
        return results



# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import contextlib
import functools
import traceback
import urllib.parse
import copy
import uuid
import re

import defusedxml.ElementTree as ET

from ..mixins import LinkedValidatorMixin, SubValidationMixin
from ..validator_logging import ValidatorLogError, ValidatorLogWarning, Path, ValidatorLog
from ..exceptions import FailFastException


class BaseValidator(LinkedValidatorMixin, SubValidationMixin):
    """Defines basic validation behaviour and expected attributes
    of any IIIF validators that inherit from it."""

    PRESENTATION_API_URI = "http://iiif.io/api/presentation/2/context.json"
    IMAGE_API_2 = "http://iiif.io/api/image/2/context.json"
    IMAGE_API_1 = "http://iiif.io/api/image/1/context.json"

    # The following constants will be iterated through and have their
    # values checked on every validation to produce warnings and errors
    # based on key constraints. Each inheritor should define these.

    #: The fields which may appear on this resource.
    KNOWN_FIELDS = {}

    #: The fields which are forbidden on this resource.
    FORBIDDEN_FIELDS = {}

    #: The fields which are required on this resource.
    REQUIRED_FIELDS = {}

    #: The fields which are recommended on this resource.
    RECOMMENDED_FIELDS = {}

    # The set of acceptable viewHints on this resource.
    VIEW_HINTS = {}

    # The set of acceptable viewDirections on this resource.
    VIEW_DIRS = {}

    # The set of fields which may appear on _any_ resource.
    COMMON_FIELDS = {
        "label", "metadata", "description", "thumbnail", "attribution", "license", "logo",
        "@id", "@type", "viewingHint", "seeAlso", "service", "related", "rendering", "within"
    }

    # The path suffixes which are allowed to contain HTML.
    HTML_ALLOWED_FIELDS = {('description',), ('attribution',),
                           ('metadata', 'value'), ('metadata', 'label'),
                           ('@value',)}

    # The attributes allowed for each html field.
    HTML_ALLOWED_ATTRIBUTES = {
        'a': {'href'},
        'img': {'src', 'alt'}
    }

    # The HTML tags which are allowed to appear in a text field.
    HTML_ALLOWED_TAGS = {'a', 'b', 'br', 'i', 'img', 'p', 'span'}

    # The HTML tags which are expressly forbidden.
    HTML_FORBIDDEN_TAGS = {'script', 'style', 'object', 'form', 'input'}

    # Catch all regex for XML in string.
    _XML_TAG_REGEX = re.compile(r'<\/?(\w+)((\s+\w+(\s*=\s*(?:\".*?\"|\'.*?\'|[\^\'\">\s]+))?)+\s*|\s*)\/?>', re.DOTALL)
    _XML_COMMENT_REGEX = re.compile(r'<!--.*?-->', re.DOTALL)
    _XML_CDATA_REGEX = re.compile(r'<!\[CDATA\[.*?\]\]>')

    def __init__(self, iiif_validator=None):
        LinkedValidatorMixin.__init__(self, iiif_validator=iiif_validator)
        SubValidationMixin.__init__(self)
        self._path = Path(tuple())
        self._json = None
        self.corrected_doc = None

        self._LangValPairs = {
            '@language': functools.partial(self._repeatable_string_type, "@language"),
            '@value': functools.partial(self._repeatable_string_type, "@value")
        }

        self._MetadataItemSchema = {
            'label': functools.partial(self._str_or_val_lang_type, "label"),
            'value': functools.partial(self._str_or_val_lang_type, "value")
        }

        self._common_fields_mapping = {
            "@id": self.id_field,
            "label": self.label_field,
            "metadata": self.metadata_field,
            "description": self.description_field,
            "thumbnail": self.thumbnail_field,
            "logo": self.logo_field,
            "attribution": self.attribution_field,
            "@type": self.type_field,
            "license": self.license_field,
            "related": self.related_field,
            "rendering": self.rendering_field,
            "service": self.service_field,
            "seeAlso": self.seeAlso_field,
            "within": self.within_field,
            'viewingHint': self.viewing_hint_field
        }

    @staticmethod
    def errors_to_warnings(fn):
        """Cast any errors to warnings on any ``*_field`` or ``*_type`` function.

        Works by patching the BaseValidator.log_error to refer to
        BaseValidator.log_warning. These methods should not be
        overridden in children.
        """
        def coerce_errors(*args, **kwargs):
            old_log_error = BaseValidator.log_error
            try:
                BaseValidator.log_error = BaseValidator.log_warning
                val = fn(*args, **kwargs)
            finally:
                BaseValidator.log_error = old_log_error
            return val

        return coerce_errors

    @staticmethod
    def warnings_to_errors(fn):
        """Cast any warnings to errors on any ``*_field`` or ``*_type`` function.

        Works by patching the BaseValidator.log_warning to refer to
        BaseValidator.log_error. These methods should not be
        overridden in children.
        """

        def coerce_warnings(*args, **kwargs):
            old_log_warning = BaseValidator.log_warning
            try:
                BaseValidator.log_warning = BaseValidator.log_error
                val = fn(*args, **kwargs)
            finally:
                BaseValidator.log_warning = old_log_warning
            return val

        return coerce_warnings

    @contextlib.contextmanager
    def _temp_path(self, path):
        """Temporarily set the path to the given path.

        Useful when validating an embedded dictionary
        and adding its path to the current path.

        :param path (Path or tuple): The path to temporarily use.
        """
        old_path = self._path
        try:
            if isinstance(path, tuple):
                self._path = Path(path)
            elif isinstance(path, Path):
                self._path = path
            else:
                raise ValueError("path must be Path or tuple.")
            yield
        finally:
            self._path = old_path

    def mute_errors(self, fn, *args, **kwargs):
        """Run given function and catch any of the errors it logged.

        The self._errors key will not be changed by using this function.

        Works by patching the BaseValidator.log_error function.
        BaseValidator.log_error should not be overridden in children.
        """
        caught_errors = set()

        def patched_log_error(self, field, msg):
            tb = traceback.extract_stack()[:-1] if self.debug else None
            caught_errors.add(ValidatorLogError(msg, self._path + (field,), tb))

        old_log_error = BaseValidator.log_error
        try:
            BaseValidator.log_error = patched_log_error
            val = fn(*args, **kwargs)
        finally:
            BaseValidator.log_error = old_log_error
        return val, caught_errors

    def _reset(self, path):
        """Reset the validator to handle a new chunk of data."""
        self._json = None
        self.is_valid = None
        self._errors = ValidatorLog(self.unique_logging)
        self._warnings = ValidatorLog(self.unique_logging)
        self._path = path

    def setup(self):
        pass

    def _validate(self, json_dict, path=None, **kwargs):
        """Entry point for callers to validate a chunk of data."""

        # Reset the validator object constants.
        if not path:
            path = Path(tuple())
        self._reset(path)

        self._json = json_dict
        try:
            val = self._run_validation(**kwargs)
            val = self._check_common_fields(val)
            self._raise_additional_warnings(val)
            self.corrected_doc = self.modify_final_return(val)
        finally:
            if self._errors:
                self.is_valid = False
            else:
                self.is_valid = True

    def _compare_dicts(self, schema, value):
        """Compare a schema to a dict.

        Emulates the behaviors of the voluptuous library, which was
        previously used. Iterates through the schema (keys), calling each
        function (values) on the corresponding entry in the `value` dict.

        :param schema: A dict where each key maps to a function.
        :param value: A dict to validate against the schema."""
        corrected = copy.copy(value)
        for key, fn in schema.items():
            if key in value:
                corrected[key] = fn(corrected[key])
        return corrected

    def _run_validation(self, **kwargs):
        """Do the actual action of validation. Called by validate()."""
        raise NotImplemented

    def _raise_additional_warnings(self, validation_results):
        """Inspect the block and raise any SHOULD warnings.

        This method is called only if the manifest validates without errors.
        It is passed the block that was just validated. This is the opportunity
        to inspect for fields which SHOULD be there and throw warnings.
        """
        pass

    def modify_final_return(self, validation_results):
        """Do any final corrections or checks on a block before it is returned.

        This method is passed whatever value the validator is about to return to
        it's caller. Here you can check for missing keys, compare neighbours,
        make modifications or additions: anything you'd like to check or correct
        before return.

        :param validation_results: A dict representing a json object.
        :return (dict): The sole argument, with some modification applied to it.
        """
        return validation_results

    def log_warning(self, field, msg):
        """Add a warning to the validator if warnings are being caught.

        This method should not be overridden in subclasses, as doing so
        is likely to break the error and warning coercion decorators.

        :param field: The field the warning was raised on.
        :param msg: The message to associate with the warning.
        """
        if self.collect_warnings:
            tb = traceback.extract_stack()[:-1] if self.debug else None
            warn = ValidatorLogWarning(msg, self._path + (field,), tb)
            if self.verbose:
                self._IIIFValidator.logger.warning(str(warn))
            self._warnings.add(warn)

    def log_error(self, field, msg):
        """Add an error to the validator.

        This method should not be overridden in subclasses, as doing so
        is likely to break the error and warning coercion decorators.

        :param field: The field the error was raised on.
        :param msg: The message to associate with the error.
        """
        if self.collect_errors:
            tb = traceback.extract_stack()[:-1] if self.debug else None
            err = ValidatorLogError(msg, self._path + (field,), tb)
            if self.verbose:
                self._IIIFValidator.logger.error(str(err))
            self._errors.add(err)
        if self.fail_fast:
            raise FailFastException

    def _check_common_fields(self, val):
        """Validate fields that could appear on any resource."""
        return self._compare_dicts(self._common_fields_mapping, val)

    def _check_recommended_fields(self, resource, r_dict, fields):
        """Log warnings if fields which should be in r_dict are not.

        :param resource (str): The name of the resource represented by r_dict
        :param r_dict (dict): The dict that will have it's keys checked.
        :param fields (list): The keys to check for in r_dict.
        """
        for f in fields:
            if not r_dict.get(f):
                self.log_warning(f, "{} SHOULD have {} field.".format(resource, f))

    def _check_unknown_fields(self, resource, r_dict, fields):
        """Log warnings if any fields which are not known in context are present.

        :param resource (str): The name of the resource represented by r_dict
        :param r_dict (dict): The dict to have its keys checked.
        :param fields (set): Known key names for this resource.
        """
        for key in r_dict.keys():
            if key not in fields:
                self.log_warning(key, "Unknown key '{}' in '{}'".format(key, resource))

    def _check_forbidden_fields(self, resource, r_dict, fields):
        """Log warnings if keys which are forbidden in context are present.

        :param resource (str): The name of the resource represented by r_dict.
        :param r_dict (dict): The dict to have its keys checked.
        :param fields (set): Forbidden key names for this resource.
        """
        for key in r_dict.keys():
            if key in fields:
                self.log_error(key, "Key '{}' is not allowed in '{}'".format(key, resource))

    def _check_required_fields(self, resource, r_dict, fields):
        """Log errors if the required fields are missing.

        :param resource (str): The name of the resource represented by r_dict.
        :param r_dict (dict): The dict to have its keys checked.
        :param fields (set): Forbidden key names for this resource.
        """
        for f in fields:
            if f not in r_dict:
                self.log_error(f, "Key '{}' is required in '{}'".format(f, resource))

    def _check_all_key_constraints(self, resource, r_dict):
        """Call all key constraint checking methods."""
        if not isinstance(r_dict, dict):
            self.log_error(resource, "'{}' must be json-object, not {}".format(resource, type(r_dict).__name__))
            return r_dict

        self._check_forbidden_fields(resource, r_dict, self.FORBIDDEN_FIELDS)
        self._check_required_fields(resource, r_dict, self.REQUIRED_FIELDS)
        self._check_recommended_fields(resource, r_dict, self.RECOMMENDED_FIELDS)
        self._check_unknown_fields(resource, r_dict, self.KNOWN_FIELDS)
        return self._check_common_fields(r_dict)

    # Field definitions #
    def _optional(self, field, fn):
        """Wrap a function to make its value optional (null and '' allows)"""

        def new_fn(*args):
            if args[0] == "" or args[0] is None:
                self.log_warning(field, "'{}' field should not be included if it is empty.".format(field))
                return args[0]
            return fn(*args)

        return new_fn

    def _not_allowed(self, field, value):
        """Raise invalid as this key is not allowed in the context."""
        self.log_error(field, "'{}' is not allowed here".format(field))
        return value

    def _str_or_val_lang_type(self, field, value):
        """Check value is str or lang/val pairs, else raise ValidatorLogError.

        Allows for repeated strings as per 5.3.2.
        """
        if isinstance(value, str):
            # Check for invalid and forbidden html.
            self._check_html(field, value)
            return value
        if isinstance(value, list):
            return [self._str_or_val_lang_type(field, val) for val in value]
        if isinstance(value, dict):
            if "@value" not in value:
                self.log_error(field, "Field has no '@value' key where one is required.")
                return value
            return self._compare_dicts(self._LangValPairs, value)
        self.log_error(field, "Illegal type (should be str, list, or dict)")
        return value

    def _repeatable_string_type(self, field, value):
        """Allows for repeated strings as per 5.3.2."""
        if isinstance(value, str):
            # Check for invalid and forbidden html.
            self._check_html(field, value)
            return value
        if isinstance(value, list):
            for val in value:
                if not isinstance(val, str):
                    self.log_error(field, "Overly nested strings: '{}'".format(value))
            return value
        self.log_error(field, "Got '{}' when expecting string or repeated string.".format(value))
        return value

    def _repeatable_uri_type(self, field, value):
        """Allow single or repeating URIs.

        Based on 5.3.2 of Presentation API
        """
        if isinstance(value, list):
            return [self._uri_type(field, val) for val in value]
        else:
            return self._uri_type(field, value)

    def _http_uri_type(self, field, value):
        """Allow single URI that MUST be http(s)

        Based on 5.3.2 of Presentation API
        """
        return self._uri_type(field, value, http=True)

    def _uri_type(self, field, value, http=False):
        """Check value is URI type or raise ValidatorLogError.

        Allows for multiple URI representations, as per 5.3.1 of the
        Presentation API.
        """
        if isinstance(value, str):
            return self._string_uri(field, value, http)
        elif isinstance(value, dict):
            emb_uri = value.get('@id')
            if not emb_uri:
                self.log_error(field, "URI not found: '{}'".format(value))
                return value
            value['@id'] = self._string_uri(field, emb_uri, http)
            return value
        else:
            self.log_error(field, "Can't parse URI: {}".format(value))
            return value

    def _string_uri(self, field, value, http=False):
        """Validate that value is a string that can be parsed as URI.

        This is the last stop on the recursive structure for URI checking.
        Should not actually be used in schema.
        """
        # Always raise invalid if the string field is not a string.
        if not isinstance(value, str):
            self.log_error(field, "URI is not string: '{}'".format(value))
            return value

        # Check for invalid and forbidden html.
        self._check_html(field, value)

        # Try to parse the url.
        try:
            pieces = urllib.parse.urlparse(value)
        except AttributeError as a:
            self.log_error(field, "URI is not valid: '{}'".format(value))
            return value
        if not all([pieces.scheme, pieces.netloc]):
            self.log_error(field, "URI is not valid: '{}'".format(value))
        if http and pieces.scheme not in ['http', 'https']:
            self.log_error(field, "URI must be http: '{}'".format(value))
        return value

    def _check_html(self, field, value):
        """Check that the value does not contain html where not allowed.

        Logs a warning if any tag not in HTML_ALLOWED_TAGS is present.
        Logs an error if any tag in HTML_FORBIDDEN_TAGS is present.
        Logs an error if any html tag is found in a field not in HTML_ALLOWED_FIELDS.
        """
        # Disregarding indices in paths, check if the suffix of the current path
        # is one which can validly contain html.
        temp_path = self._path + field
        field_allowed_html = any(temp_path.no_index_endswith(x) for x in self.HTML_ALLOWED_FIELDS)

        # Bool marking if this field contains valid xml markup.
        field_is_valid_xml = False

        # Bool marking if this field contains any tags.
        field_contains_tags = bool(self._XML_TAG_REGEX.search(value))

        # Bail if tags detected but first char is not '<'
        if field_contains_tags and value[0] != "<":
            self.log_error(field, "If field contains HTML, it must start with character '<'.")
            return

        # Error and exit if XML comments are detected.
        field_contains_comments = bool(self._XML_COMMENT_REGEX.search(value))
        if field_contains_comments:
            self.log_error(field, "XML comments not allowed.")
            return

        # Error and exit if CDATA sections are detected.
        field_contains_cdata = bool(self._XML_CDATA_REGEX.search(value))
        if field_contains_cdata:
            self.log_error(field, "CDATA sections not allowed.")
            return

        # Try to parse the field and record if the field is valid xml.
        if field_contains_tags:
            try:
                et = ET.fromstring(value)
                field_is_valid_xml = True
            except ET.ParseError:
                field_is_valid_xml = False

        # Return now if no tags are found.
        if not field_contains_tags:
            return

        # Log error and return if this field is not allowed to have HTML in it.
        if (field_is_valid_xml or field_contains_tags) and not field_allowed_html:
            self.log_error(field, "HTML not allowed in this field.")
            return

        # Log error and return if the HTML is malformed in some way.
        if field_contains_tags and not field_is_valid_xml:
            self.log_error(field, "Field contains tags but is not valid HTML.")
            return

        def check_html_element(elem):
            """Recursively validate elements in etree."""
            tag, attributes = elem.tag, elem.attrib.keys()

            # Log error and return if tag is forbidden.
            if tag in self.HTML_FORBIDDEN_TAGS:
                self.log_error(field, "Forbidden tag '<{}>' in html.".format(tag))
                return False

            # Log error and return if forbidden attributes are present.
            allowed_attributes = self.HTML_ALLOWED_ATTRIBUTES.get(tag, set())
            for attr in attributes:
                if attr not in allowed_attributes:
                    self.log_error(field, "HTML tag '<{}>' not allowed attribute '{}'.".format(tag, attr))
                    return False

            # Log warning if tag is not explicitly mentioned as being safe.
            if tag not in self.HTML_ALLOWED_TAGS:
                self.log_warning(field, "HTML tag '<{}>' of uncertain validity "
                                        "(valid tags are <a>, <b>, <br>, <i>, <img>, <p>, and <span>)".format(tag))

            for child_elem in elem:
                child_valid = check_html_element(child_elem)
                if not child_valid:
                    return False
            return True

        check_html_element(et)

    # Common field definitions.
    def id_field(self, value):
        """Validate the ``@id`` field of the resource."""
        if value.startswith("urn:uuid:"):
            id_uuid = value.replace("urn:uuid:", "")
            try:
                uuid.UUID(id_uuid)
            except ValueError:
                self.log_error("@id", "Invalid UUID in @id.")
            finally:
                return value
        return self._http_uri_type("@id", value)

    def type_field(self, value):
        """Validate the ``@type`` field of the resource."""
        raise NotImplemented

    def label_field(self, value):
        """Validate the ``label`` field of the resource."""
        return self._str_or_val_lang_type("label", value)

    def description_field(self, value):
        """Validate the ``description`` field of the resource."""
        return self._str_or_val_lang_type("description", value)

    def attribution_field(self, value):
        """Validate the ``attribution`` field of the resource."""
        return self._str_or_val_lang_type("attribution", value)

    def license_field(self, value):
        """Validate the ``license`` field of the resource."""
        return self._repeatable_uri_type("license", value)

    def related_field(self, value):
        """Validate the ``related`` field of the resource."""
        return self._repeatable_uri_type("related", value)

    def rendering_field(self, value):
        """Validate the ``rendering`` field of the resource."""
        return self._repeatable_uri_type("rendering", value)

    def service_field(self, value):
        """Validate the ``service`` field of the resource."""
        return self._repeatable_uri_type("service", value)

    def seeAlso_field(self, value):
        """Validate the ``seeAlso`` field of the resource."""
        return self._repeatable_uri_type("seeAlso", value)

    def within_field(self, value):
        """Validate the ``within`` field of the resource."""
        return self._repeatable_uri_type("within", value)

    def height_field(self, value):
        """Validate ``height`` field."""
        if not isinstance(value, int):
            self.log_error("height", "height must be int.")
        return value

    def width_field(self, value):
        """Validate ``width`` field."""
        if not isinstance(value, int):
            self.log_error("width", "width must be int.")
        return value

    def metadata_field(self, value):
        """Validate the `metadata` field of the resource.

        Recurse into keys/values and checks that they are properly formatted.
        """
        if not isinstance(value, list):
            self.log_error("metadata", "Metadata MUST be a list")
            return value

        result = []
        with self._temp_path(self._path + ("metadata",)):
            for i, m in enumerate(value):
                with self._temp_path(self._path + i):
                    result.append(self._metadata_entry(m))
        return result

    def _metadata_entry(self, value):
        if not isinstance(value, dict):
            self.log_error("value", "Entries must be dictionaries.")
            return value
        if "label" not in value:
            self.log_error("label", "metadata entries must have labels.")
            return value
        elif "value" not in value:
            self.log_error("value", "metadata entries must have values")
            return value
        else:
            return {
                'label': self._str_or_val_lang_type("label", value.get("label")),
                'value': self._str_or_val_lang_type("value", value.get("value"))
            }

    def thumbnail_field(self, value):
        """Validate the ``thumbnail`` field of the resource."""
        return self._general_image_resource("thumbnail", value)

    def logo_field(self, value):
        """Validate the ``logo`` field of the resource."""
        return self._general_image_resource("logo", value)

    def _general_image_resource(self, field, value):
        """Image resource validator for logos and thumbnails. Basic logic is:

        -Check if field is string. If yes, warn that IIIF image service is preferred.
        -If a IIIF image service is avaliable,  try to validate it.
        -Otherwise, check that it's ID is at least a uri.
        """

        if isinstance(value, str):
            self.log_warning(field, "{} SHOULD be IIIF image service.".format(field))
            return self._uri_type(field, value)
        if isinstance(value, dict):
            service = value.get("service")
            if service and service.get("@context") == "http://iiif.io/api/image/2/context.json":
                value['service'] = self.ImageContentValidator.service_field(service)
                return value
            else:
                val = self._uri_type(field, value)
                self.log_warning(field, "{} SHOULD be IIIF image service.".format(field))
                return val
        self.log_error(field, "{} type should be string or dict.".format(field))
        return value

    def viewing_hint_field(self, value):
        """Validate ``viewingHint`` field against ``VIEW_HINTS`` set."""
        if value not in self.VIEW_HINTS:
            val, errors = self.mute_errors(self._uri_type, "viewingHint", value)
            if errors:
                self.log_error("viewingHint", "viewingHint '{}' is not valid and not uri.".format(value))
        return value

    def viewing_dir_field(self, value):
        """Validate ``viewingDir`` field against ``VIEW_DIRS`` set."""
        if value not in self.VIEW_DIRS:
            self.log_error("viewingDirection", "viewingDirection '{}' is not valid and not uri.".format(value))
        return value



try:
    import ujson as json
except ImportError:
    import json

import os


from .validator_testing_tools import ValidatorTestingTools
from tripoli import IIIFValidator


class TestIIIFValidator(ValidatorTestingTools):
    def setUp(self):
        self.test_subject = IIIFValidator()
        self.test_subject.logger.setLevel("CRITICAL")
        self.test_subject.fail_fast = False
        self.test_subject.collect_errors = True
        self.test_subject.collect_warnings = True
        self.test_subject.debug = True

        self.base_dir = os.path.dirname(os.path.realpath(__file__))

    def test_valid_manifest(self):
        """Test that a valid manifest raises no errors."""
        with open(os.path.join(self.base_dir, 'fixtures/valid_manifest')) as f:
            man = json.load(f)
            self.test_subject.validate(man)
        self.assertFalse(self.has_errors())

    def test_text_manifest(self):
        """Test that a manifest can be passed as text."""
        with open(os.path.join(self.base_dir, 'fixtures/valid_manifest')) as f:
            self.test_subject.validate(f.read())
        self.assertFalse(self.has_errors())




# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json
import logging

from .exceptions import FailFastException, TypeParseException
from .mixins import SubValidationMixin
from .validator_logging import ValidatorLogError, ValidatorLog, Path
from .resource_validators import (
    ManifestValidator, SequenceValidator, CanvasValidator,
    ImageContentValidator, AnnotationValidator)

__version__ = "1.1.3"


class IIIFValidator(SubValidationMixin):
    def __init__(self):
        super().__init__()
        self._ManifestValidator = None
        self._AnnotationValidator = None
        self._CanvasValidator = None
        self._SequenceValidator = None
        self._ImageContentValidator = None

        #: ``logging.getLogger()`` used to print output.
        self.logger = logging.getLogger("tripoli")

        #: Sets whether or not to save tracebacks in warnings/errors.
        #: Default ``False``.
        self.debug = False

        #: Sets whether or not warnings are logged.
        #: Default ``True``.
        self.collect_warnings = True

        #: Sets whether or not errors are logged.
        #: Default ``True``.
        self.collect_errors = True

        #: When ``True``, validation stops at first error hit (faster).
        #: If ``False``, entire document will always be validated.
        #: Default ``True``.
        #:
        #: Note: Turning ``fail_fast`` off may cause the validator to raise
        #: unexpected exceptions if the the document is grossly invalid
        #: (for instance, if an integer is supplied where a list is expected).
        self.fail_fast = True

        #: If corrections were made during validation, the corrected document
        #: will be placed here.
        self.corrected_doc = {}

        #: If ``True``, prints all errors and warnings as they occur.
        #: If ``False``, errors and warnings only printed after de-duplication.
        #: Default ``False``.
        self.verbose = False

        #: If ``True``, only one instance of duplicate logged messages will be saved.
        #: If ``False``, all logged messages will be saved.
        #: Default ``True``.
        #:
        #: Example: If set to true, then if every canvas has error A, instead
        #: of having the errors (Error(A, canvas[0]), Error(A, canvas[1]), ...), you
        #: will only get Error(A, canvas[0]) (the first error of type A on a canvas).
        self.unique_logging = True

        self._setup_to_validate()

    @property
    def ManifestValidator(self):
        """An instance of a ManifestValidator."""
        return self._ManifestValidator

    @property
    def SequenceValidator(self):
        """An instance of a SequenceValidator."""
        return self._SequenceValidator

    @property
    def CanvasValidator(self):
        """An instance of a CanvasValidator."""
        return self._CanvasValidator

    @property
    def AnnotationValidator(self):
        """An instance of an AnnotationValidator"""
        return self._AnnotationValidator

    @property
    def ImageContentValidator(self):
        """An instance of an ImageContentValidator"""
        return self._ImageContentValidator

    @ManifestValidator.setter
    def ManifestValidator(self, value):
        self._ManifestValidator = value(self)

    @SequenceValidator.setter
    def SequenceValidator(self, value):
        self._SequenceValidator = value(self)

    @CanvasValidator.setter
    def CanvasValidator(self, value):
        self._CanvasValidator = value(self)

    @AnnotationValidator.setter
    def AnnotationValidator(self, value):
        self._AnnotationValidator = value(self)

    @ImageContentValidator.setter
    def ImageContentValidator(self, value):
        self._ImageContentValidator = value(self)

    def _setup_to_validate(self):
        """Make sure all links to sub validators exist."""
        if not self._ManifestValidator:
            self._ManifestValidator = ManifestValidator(self)
        if not self._AnnotationValidator:
            self._AnnotationValidator = AnnotationValidator(self)
        if not self._CanvasValidator:
            self._CanvasValidator = CanvasValidator(self)
        if not self._SequenceValidator:
            self._SequenceValidator = SequenceValidator(self)
        if not self._ImageContentValidator:
            self._ImageContentValidator = ImageContentValidator(self)

        self._TYPE_MAP = {
            "sc:Manifest": self._ManifestValidator,
            "sc:Sequence": self._SequenceValidator,
            "sc:Canvas": self._CanvasValidator,
            "oa:Annotation": self._AnnotationValidator
        }
        self._errors = ValidatorLog(self.unique_logging)
        self._warnings = ValidatorLog(self.unique_logging)
        self.corrected_doc = {}

    def _set_from_sub(self, sub):
        """Set the validation attributes to those of a sub_validator.

        Called after sub_validate'ing with validator sub.

        :param sub: A BaseValidator implementing Validator.
        """
        self.is_valid = sub.is_valid
        self.corrected_doc = sub.corrected_doc

    def _output_logging(self):
        """Sends errors and warnings to the logger."""
        for err in self.errors:
            self.logger.error(err.log_str())
        for warn in self.warnings:
            self.logger.warning(warn.log_str())

    def _parse_json(self, json_dict):
        if isinstance(json_dict, str):
            try:
                json_dict = json.loads(json_dict)
            except ValueError:
                self._exit_early("Could not parse json.")
        return json_dict

    def _get_validator(self, json_dict):
        """Parse json_dict and return the correct validator.

        Raises a TypeParseException if this cannot be done.
        """

        doc_type = json_dict.get("@type")
        if not doc_type:
            self._exit_early("Resource has no @type.")

        validator = self._TYPE_MAP.get(doc_type)
        if not validator:
            self._exit_early("Unknown @type: '{}'".format(doc_type))
        return validator

    def _exit_early(self, msg):
        """Log an error with message, set is_valid to false, and raise TypeParseException."""
        self._errors.add(ValidatorLogError(msg, Path()))
        self.is_valid = False
        raise TypeParseException

    def validate(self, json_dict, **kwargs):
        """Determine the correct validator and validate a resource.

        :param json_dict: A dict or str of a json resource.
        """
        self._setup_to_validate()
        try:
            json_dict = self._parse_json(json_dict)
            validator = self._get_validator(json_dict)
        except TypeParseException:
            self._output_logging()
            return

        try:
            self._sub_validate(validator, json_dict, path=None, **kwargs)
        except FailFastException:
            pass
        self._set_from_sub(validator)
        self._output_logging()



# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from collections import OrderedDict

from .base_validator import BaseValidator


class ManifestValidator(BaseValidator):
    VIEW_DIRS = ['left-to-right', 'right-to-left',
                 'top-to-bottom', 'bottom-to-top']
    VIEW_HINTS = ['individuals', 'paged', 'continuous']

    KNOWN_FIELDS = BaseValidator.COMMON_FIELDS | {"viewingDirection", "navDate", "sequences", "structures", "@context"}
    FORBIDDEN_FIELDS = {"format", "height", "width", "startCanvas", "first", "last", "total", "next", "prev",
                        "startIndex", "collections", "manifests", "members", "canvases", "resources", "otherContent",
                        "images", "ranges"}
    REQUIRED_FIELDS = {"label", "@context", "@id", "@type", "sequences"}
    RECOMMENDED_FIELDS = {"metadata", "description", "thumbnail"}

    def __init__(self, iiif_validator):
        super().__init__(iiif_validator)
        self.ManifestSchema = OrderedDict((
            ('@context', self.context_field),
            ('structures', self.structures_field),
            ('sequences', self.sequences_field),
            ('viewingDirection', self.viewing_dir_field),
        ))
        self.setup()

    def _run_validation(self, **kwargs):
        self._check_all_key_constraints("manifest", self._json)
        return self._compare_dicts(self.ManifestSchema, self._json)

    def type_field(self, value):
        """Assert that ``@type`` == ``sc:Manifest``. """
        if not value == 'sc:Manifest':
            self.log_error("@type", "@type must be 'sc:Manifest'.")
        return value

    def context_field(self, value):
        """Assert that ``@context`` is the IIIF 2.0 presentation API."""
        if isinstance(value, str):
            if not value == self.PRESENTATION_API_URI:
                self.log_error("@context", "'@context' must be set to '{}'".format(self.PRESENTATION_API_URI))
        if isinstance(value, list):
            if self.PRESENTATION_API_URI not in value:
                self.log_error("@context", "'@context' must be set to '{}'".format(self.PRESENTATION_API_URI))
        return value

    def structures_field(self, value):
        """Validate the ``structures`` field."""
        return value

    def sequences_field(self, value):
        """Validate ``sequences`` list for Manifest.

        Checks that exactly 1 sequence is embedded.
        """
        if not isinstance(value, list):
            self.log_error("sequences", "'sequences' MUST be a list")
            return value

        if len(value) == 0:
            self.log_error("sequences", "Manifest requires at least one sequence")
            return value

        results = []
        path = self._path + ("sequences",)
        for i, seq in enumerate(value):
            temp_path = path + i
            if i == 0:
                results.append(self._sub_validate(self.SequenceValidator, seq, temp_path, emb=True))
            else:
                results.append(self._sub_validate(self.SequenceValidator, seq, temp_path, emb=False))
        return results



from .validator_logging import ValidatorLog

class LinkedValidatorMixin:
    """Basic support for storing 'global' references in a single administrative class."""

    def __init__(self, iiif_validator=None):
        #: The IIIFValidator containing this validator.
        self._IIIFValidator = iiif_validator

    @property
    def collect_warnings(self):
        return self._IIIFValidator.collect_warnings

    @property
    def collect_errors(self):
        return self._IIIFValidator.collect_errors

    @property
    def debug(self):
        return self._IIIFValidator.debug

    @property
    def fail_fast(self):
        return self._IIIFValidator.fail_fast

    @property
    def verbose(self):
        return self._IIIFValidator.verbose

    @property
    def unique_logging(self):
        return self._IIIFValidator.unique_logging

    @property
    def ManifestValidator(self):
        return self._IIIFValidator._ManifestValidator

    @property
    def SequenceValidator(self):
        return self._IIIFValidator._SequenceValidator

    @property
    def CanvasValidator(self):
        return self._IIIFValidator._CanvasValidator

    @property
    def AnnotationValidator(self):
        return self._IIIFValidator._AnnotationValidator

    @property
    def ImageContentValidator(self):
        return self._IIIFValidator._ImageContentValidator


class SubValidationMixin:
    """Provides needed parts to accumulate errors and delegate validation."""

    def __init__(self):
        self._errors = ValidatorLog()
        self._warnings = ValidatorLog()
        self.is_valid = None

    @property
    def errors(self):
        """A list of ValidatorLogError from the previous call to validate()."""
        return list(self._errors)

    @property
    def warnings(self):
        """A list of ValidatorLogWarnings from the previous call to validate()."""
        return list(self._warnings)

    def print_errors(self):
        """Print accumulated errors in a nice format."""
        for err in self._errors:
            print(err)

    def print_warnings(self):
        """Print accumulated warnings in a nice format."""
        for warn in self._warnings:
            print(warn)

    def _sub_validate(self, subschema, value, path, **kwargs):
        """Validate a field using another Validator.

        :param subschema: A BaseValidator implementing object.
        :param value (dict): The data to be validated.
        :param path (tuple): The path where the above data exists.
            Example: ('sequences', 'canvases') for the CanvasValidator.
        :param kwargs: Any keys to subschema._run_validation()
            - canvas_uri: String passed to AnnotationValidator from
              CanvasValidator to ensure 'on' key is valid.
            - raise_warnings: bool to decide if warnings will be recorded
              or not.
        """
        try:
            subschema._validate(value, path, **kwargs)
        finally:
            if subschema._errors:
                self._errors.update(subschema._errors)
            if subschema._warnings:
                self._warnings.update(subschema._warnings)
        if subschema.corrected_doc:
            return subschema.corrected_doc
        else:
            return subschema._json



# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from collections import OrderedDict

from .base_validator import BaseValidator


class ImageContentValidator(BaseValidator):
    KNOWN_FIELDS = BaseValidator.COMMON_FIELDS | {"@context", "height", "width", "format"}
    FORBIDDEN_FIELDS = {"viewingDirection", "navDate", "startCanvas", "first", "last", "total",
                        "next", "prev", "startIndex", "collections", "manifests", "members",
                        "sequences", "structures", "canvases", "resources", "otherContent",
                        "images", "ranges"}
    REQUIRED_FIELDS = {'@type', '@id'}

    def __init__(self, iiif_validator):
        super().__init__(iiif_validator)
        self.ImageContentSchema = OrderedDict((
            ('@id', self.id_field),
            ('@type', self.type_field),
            ('height', self.height_field),
            ('width', self.width_field),
            ('service', self.service_field)
        ))
        self.setup()

    def _run_validation(self, **kwargs):
        self._check_all_key_constraints("resource", self._json)
        return self._compare_dicts(self.ImageContentSchema, self._json)

    def type_field(self, value):
        """Warn if ``@type != 'dctypes:Image'``"""
        if value != 'dctypes:Image':
            self.log_error('@type', "@type MUST be \'dctypes:Image\'")
        return value

    def service_field(self, value):
        """Validate the image service in this resource."""
        with self._temp_path(self._path + ('service',)):
            self._check_required_fields("image service", value, ['@id', '@context'])
            self._check_recommended_fields("image service", value, ['profile'])
            context = value.get("@context")
            if context and context != self.IMAGE_API_2:
                if context != self.IMAGE_API_1:
                    self.log_error('@context', "Must reference IIIF image API.")
                else:
                    self.log_warning('@context', "SHOULD upgrade to 2.0 IIIF image service.")
            return value



# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from collections import OrderedDict

from .base_validator import BaseValidator


class AnnotationValidator(BaseValidator):
    KNOWN_FIELDS = BaseValidator.COMMON_FIELDS | {"motivation", "resource", "on", "@context"}
    FORBIDDEN_FIELDS = {"format", "height", "width", "viewingDirection", "navDate", "startCanvas", "first",
                        "last", "total", "next", "prev", "startIndex", "collections", "manifests", "members",
                        "sequences", "structures", "canvases", "resources", "otherContent", "images", "ranges"}
    REQUIRED_FIELDS = {"@type", "on", "motivation", "resource"}

    def __init__(self, iiif_validator):
        super().__init__(iiif_validator)
        self.ImageSchema = OrderedDict((
            ("@id", self.id_field),
            ('@type', self.type_field),
            ('motivation', self.motivation_field),
            ("on", self.on_field),
            ('height', self.height_field),
            ('width', self.width_field),
            ('resource', self.resource_field)
        ))

        self.canvas_uri = None
        self.setup()

    def _raise_additional_warnings(self, validation_results):
        pass

    def _run_validation(self, canvas_uri=None, **kwargs):
        self.canvas_uri = canvas_uri
        self._check_all_key_constraints("annotation", self._json)
        return self._compare_dicts(self.ImageSchema, self._json)

    def type_field(self, value):
        """Assert that ``@type == 'oa:Annotation'``."""
        if value != "oa:Annotation":
            self.log_error("@type", "@type must be 'oa:Annotation'.")
        return value

    def motivation_field(self, value):
        """Assert that ``motivation == 'sc:painting'``."""
        if value != "sc:painting":
            self.log_error("motivation", "motivation must be 'sc:painting'.")
        return value

    def on_field(self, value):
        """Validate the ``on`` field."""
        if self.canvas_uri and value != self.canvas_uri:
            self.log_error("on", "'on' must reference the canvas URI.")
        return value

    def resource_field(self, value):
        """Validate ``resources`` list.

        Calls a sub-validation procedure handled by the :class:`ImageContentValidator`.
        """
        path = self._path + ("resource", )
        return self._sub_validate(self.ImageContentValidator, value, path)



# Copyright (c) 2016 Alex Parmentier, Andrew Hankinson

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import sys
from setuptools import setup, find_packages
from codecs import open

from tripoli.tripoli import __version__

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist bdist_wheel upload')
    sys.exit()

with open('README.rst', 'r', 'utf-8') as f:
    read_me = f.read()

with open('HISTORY.rst', 'r', 'utf-8') as f:
    history = f.read()

setup(
    name='tripoli',
    packages=find_packages(),
    version=__version__,
    license='https://opensource.org/licenses/MIT',
    description='IIIF document validation.',
    long_description=read_me + "\n\n" + history,
    author='Alex Parmentier',
    author_email='a.g.parmentier@gmail.com',
    url='https://github.com/DDMAL/tripoli',
    download_url='https://github.com/DDMAL/tripoli/tarball/master',
    keywords=['validator', 'IIIF'],
    install_requires=['defusedxml'],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Multimedia :: Graphics :: Presentation",
        "Intended Audience :: Developers",
        "Environment :: Web Environment",
        "Programming Language :: Python"],
    test_suite='tests'
)
